/*
       B R O C C O L I  --  The Bro Client Communications Library

Copyright (C) 2004 Christian Kreibich <christian.kreibich(at)cl.cam.ac.uk>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include <errno.h>

#ifdef __MINGW32__
#include <winsock.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#endif

#include <bro_types.h>
#include <bro_debug.h>
#include <bro_buf.h>
#include <bro_event.h>
#include <bro_event_reg.h>
#include <bro_util.h>
#include <bro_type.h>
#include <bro_val.h>
#include <bro_io.h>
#include <bro_id.h>
#ifdef BRO_PCAP_SUPPORT
#include <bro_packet.h>
#endif

/* A static array that can translate message types into strings so
 * it's clearer what's going on when debugging:
 */
static char * msg_type_str[] = {
  "BRO_MSG_NONE",
  "BRO_MSG_VERSION",
  "BRO_MSG_SERIAL",
  "BRO_MSG_CLOSE",
  "BRO_MSG_CLOSE_ALL",
  "BRO_MSG_ERROR",
  "BRO_MSG_CONNECTTO",
  "BRO_MSG_CONNECTED",
  "BRO_MSG_REQUEST",
  "BRO_MSG_LISTEN",
  "BRO_MSG_LISTEN_STOP",
  "BRO_MSG_STATS",
  "BRO_MSG_CAPTURE_FILTER",
  "BRO_MSG_REQUEST_SYNC",
  "BRO_MSG_HANDSHAKE_COMPLETE",
};

static const char *
msg_type_2_str(int type)
{
  if (type < 0 || type >= BRO_MSG_MAX)
    return "<invalid>";

  return msg_type_str[type];
}

/* The maximum number of messages we queue before we start
 * dropping messages. Might be worth moving this to the
 * config file.
 */
#define BRO_MSG_QUEUELEN_MAX      1000

/* The maximum number of attempts we make in one go in order
 * to extract data from the pipe. It appears that OpenSSL does
 * not fully hide the record-oriented transfer and repeated
 * writes at one end also require repeated reads at the other
 * end. If anyone can confirm or deny this, please get in touch.
 */
#define BRO_BIO_READ_ROUNDS_MAX   20

/* Reading-related --------------------------------------------------- */

static int
io_read_chunk_size(BroConn *bc, uint32 *chunk_size)
{
  if (! __bro_buf_ptr_read(&bc->rx_buf, chunk_size, sizeof(uint32)))
    {
      D(("Couldn't read chunk size\n"));
      return FALSE;
    }

  *chunk_size = ntohl(*chunk_size);

  if (! __bro_buf_ptr_check(&bc->rx_buf, *chunk_size))
    {
      D(("Not all chunk data available\n"));
      return FALSE;
    }

  D(("We have a chunk of size %i\n", *chunk_size));
  return TRUE;
}

static int
io_read_msg_hdr(BroConn *bc, BroMsgHeader *msg_hdr)
{
  if (! __bro_buf_ptr_read(&bc->rx_buf, msg_hdr, sizeof(BroMsgHeader)))
    {
      D(("Couldn't read message header\n"));
      return FALSE;
    }
  
  msg_hdr->hdr_peer_id = ntohs(msg_hdr->hdr_peer_id);
  
  return TRUE;
}

static int
io_input_buffer_fill(BroConn *bc)
{
  BroBuf tmp_buf;
  int i, n, total = 0;
  
  for (i = 0; i < BRO_BIO_READ_ROUNDS_MAX; i++)
    {
      __bro_buf_init(&tmp_buf);
      n = __bro_openssl_read(bc, __bro_buf_get(&tmp_buf),
			     __bro_buf_get_size(&tmp_buf));
      
      if (n <= 0)
	{
	  __bro_buf_cleanup(&tmp_buf);      	  
	  
	  if (n < 0)
	    {
	      D(("Read returned error after %i/%i rounds and %i bytes\n",
		 i, BRO_BIO_READ_ROUNDS_MAX, total));
	      return -1;
	    }
	  
	  D(("Read %i bytes in %i/%i rounds.\n",
	     total, i, BRO_BIO_READ_ROUNDS_MAX));
	  
	  return total;
	}
      
      __bro_buf_append(&bc->rx_buf, __bro_buf_get(&tmp_buf), n);
      __bro_buf_cleanup(&tmp_buf);
      total += n;
    }
  
  D(("Read %i bytes in %i/%i rounds.\n",
     total, i, BRO_BIO_READ_ROUNDS_MAX));
  
  return total;
}


static int
io_skip_chunk(BroBuf *buf, uint32 buf_off, uint32 chunk_size)
{
  if (! buf)
    return FALSE;

  if (! __bro_buf_ptr_seek(buf, buf_off, SEEK_SET))
    return FALSE;

  /* We skip a uint32 for the chunk size and then chunk_size
   * bytes.
   */
  if (! __bro_buf_ptr_seek(buf, sizeof(uint32) + chunk_size, SEEK_CUR))
    return FALSE;
  
  D(("Skipping %i + %i\n", sizeof(uint32), chunk_size));

  return TRUE;
}


static int
io_process_serialization(BroConn *bc)
{
  BroVal vbuf;

  D_ENTER;

  if (! __bro_buf_read_char(&bc->rx_buf, &vbuf.val_char))
    {
      D(("Couldn't read serialization type\n"));
      D_RETURN_(FALSE);
    }

  switch (vbuf.val_char)
    {
    case 'e':
      {
	BroEvent *ev;

	D(("Processing serialized event.\n"));
	if (! (ev = __bro_event_unserialize(bc)))
	  D_RETURN_(FALSE);
	
	__bro_event_reg_dispatch(bc, ev);
	__bro_event_free(ev);      
      }
      break;

    case 'i':
      {
	BroID *id;

	D(("Processing serialized ID.\n"));
	if (! (id = (BroID *) __bro_sobject_unserialize(SER_IS_ID, bc)))
	  D_RETURN_(FALSE);
	
	D(("ID read successfully.\n"));
      }
      break;

    case 'p':
      {
#ifdef BRO_PCAP_SUPPORT
	BroPacket *packet;
	
	if (! (packet = __bro_packet_unserialize(bc)))
	  D_RETURN_(FALSE);
	
	D(("Packet read successfully.\n"));
	bro_packet_free(packet);
#else
	D_RETURN_(FALSE);
#endif
      }
      break;

    default:
      /* We do not handle anything else yet -- just say
       * we are happy and return.
       */
      D(("Unknown serialization of type %c\n", vbuf.val_char));
      D_RETURN_(FALSE);
    }

  D_RETURN_(TRUE);
}


/* Writing-related --------------------------------------------------- */

/* FIXME -- these should check return value */

static void
io_fill_raw(BroBuf *buf, BroBuf *data)
{
  __bro_buf_write_data(buf, __bro_buf_get(data), __bro_buf_get_used_size(data));
}


static void
io_fill_request(BroBuf *buf, BroRequest *req)
{
  __bro_buf_write_data(buf, req->req_dat, req->req_len);
}

static void
io_fill_msg_header(BroBuf *buf, BroMsgHeader *mh)
{
  mh->hdr_peer_id = htonl(mh->hdr_peer_id);
  __bro_buf_write_data(buf, mh, sizeof(BroMsgHeader));
}

static int      
io_msg_write(BroConn *bc, BroMsg *msg)
{
  int n, result = TRUE;
  int todo;

  D_ENTER;

  if (!bc || !msg)
    {
      D(("Input error.\n"));
      D_RETURN_(FALSE);
    }

  if (bc->tx_dead)
    {
      if (! (bc->conn_flags & BRO_CFLAG_RECONNECT))
	D_RETURN_(FALSE);
      
      bro_reconnect(bc);      

      /* We return now no matter whether we succeeded in reconnecting
       * or not, because the reconnect may have put new messages at the
       * top of the output queue that need to be processed first.
       * By returning FALSE we ensure that the given message is put last
       * in the queue and does not get sent right away.
       */
      D_RETURN_(FALSE);     
    }
  
  D((">>> Attempting write of %s\n", msg_type_2_str(msg->msg_header.hdr_type)));

  /* We will collect the message chunk in the connection's tx buffer.
   * We append stuff to it as we go along and at the end write it out.
   * When being sent, he buffer has the amount of octets to send at
   * the beginning, so the reader knows how much is coming.
   */
  __bro_buf_init(&bc->tx_buf);

  msg->msg_header_size = sizeof(BroMsgHeader);
  
  if (! __bro_buf_write_int(&bc->tx_buf, msg->msg_header_size))
    D_RETURN_(FALSE);
  
  /* Hook in the Bro message header */
  io_fill_msg_header(&bc->tx_buf, &msg->msg_header);
  
  if (msg->msg_cont_type != BRO_MSG_CONT_NONE)
    {
      uint32 msg_size_pos, msg_size_end;

      /* This starts another chunk of data (in Bro protocol speak),
       * but here we cannot yet know how big the chunk will be.
       * We save the offset in the buffer and return to it later,
       * overwriting the value with the then correct one.
       */
      msg_size_pos = __bro_buf_get_used_size(&bc->tx_buf);
      if (! __bro_buf_write_int(&bc->tx_buf, msg->msg_size))
	{
	  __bro_buf_cleanup(&bc->tx_buf);
	  D_RETURN_(FALSE);
	}
      
      /* Gather the payload of the message we are about
       * to send into the buffer BUF.
       */
      switch (msg->msg_cont_type)
	{
	case BRO_MSG_CONT_RAW:
	  D(("Filling raw data into buffer\n"));
	  io_fill_raw(&bc->tx_buf, msg->msg_cont_raw);
	  break;
	  
	case BRO_MSG_CONT_EVENT:
	  /* Check if the peer actually requested the event, and if not,
	   * drop it silently (i.e., still return success).
	   */
	  if (! __bro_ht_get(bc->ev_mask, msg->msg_cont_ev->name.str_val))
	    {
	      D(("Event '%s' not requested by peer -- dropping.\n",
		 msg->msg_cont_ev->name.str_val));
	      __bro_buf_cleanup(&bc->tx_buf);
	      D_RETURN_(TRUE);
	    }
	  
	  D(("Filling event into buffer\n"));
	  
	  if (! __bro_event_serialize(msg->msg_cont_ev, bc))
	    {
	      D(("Error during serialization.\n"));
	      __bro_buf_cleanup(&bc->tx_buf);
	      D_RETURN_(FALSE);
	    }
	  break;
	  
	case BRO_MSG_CONT_REQUEST:
	  D(("Filling request into buffer\n"));
	  io_fill_request(&bc->tx_buf, msg->msg_cont_req);
	  break;
#ifdef BRO_PCAP_SUPPORT	  
	case BRO_MSG_CONT_PACKET:
	  if (! __bro_packet_serialize(msg->msg_cont_packet, bc))
	    D_RETURN_(FALSE);
	  break;
#endif
	default:
	  D(("ERROR -- invalid message content code %i\n", msg->msg_cont_type));
	  break;
	}
      
      /* Now calculate length of entire transmission --
       * we know where we wrote the uint32 containing the
       * size of the chunk, and we know where we are now,
       * so the length is their difference, minus the uint32
       * itself.
       */
      msg_size_end = __bro_buf_get_used_size(&bc->tx_buf);
      msg->msg_size = msg_size_end - msg_size_pos - sizeof(uint32);
      D(("Sending message sized %i bytes.\n", msg->msg_size));
      
      if (! __bro_buf_ptr_seek(&bc->tx_buf, msg_size_pos, SEEK_SET))
	{
	  D(("Cannot seek to position %u -- we're screwed.\n", msg_size_pos));
	  __bro_buf_cleanup(&bc->tx_buf);
	  D_RETURN_(FALSE);
	}
      
      if (! __bro_buf_write_int(&bc->tx_buf, msg->msg_size))
	{
	  __bro_buf_cleanup(&bc->tx_buf);
	  D_RETURN_(FALSE);
	}
    }

  /* Now try to send the thing off ... */
  todo = __bro_buf_get_used_size(&bc->tx_buf);
  n    = __bro_openssl_write(bc, __bro_buf_get(&bc->tx_buf), todo);
  
  if (n <= 0)
    result = FALSE;
  
  D(("<<< Wrote %i/%i bytes of %s(%i): %i\n", n, todo,
     msg_type_2_str(msg->msg_header.hdr_type), msg->msg_num, result));
  __bro_buf_cleanup(&bc->tx_buf);
  
  D_RETURN_(result);
}


static int
io_msg_queue(BroConn *bc, BroMsg *msg)
{
  D_ENTER;

  if (!bc || !msg)
    D_RETURN_(FALSE);

  /* If the queue is empty, try to send right away.
   * If not, enqueue the event, and try to flush.
   */
  D(("Processing msg of type %s\n", msg_type_2_str(msg->msg_header.hdr_type)));

  if (! bc->msg_queue.tqh_first)
    {
      if (io_msg_write(bc, msg))
	{
	  __bro_io_msg_free(msg);
	  D(("Message sent.\n"));
	  D_RETURN_(TRUE);
	}
    }
  
  if (bc->tx_dead && ! (bc->conn_flags & BRO_CFLAG_ALWAYS_QUEUE))
    {
      D(("Connection %p disconnected, and no queuing requested: dropping message.\n", bc));
      __bro_io_msg_free(msg);
      D_RETURN_(FALSE);
    }

  TAILQ_INSERT_TAIL(&bc->msg_queue, msg, msg_queue);
  bc->msg_queue_len++;
  D(("Queue length now %i\n", bc->msg_queue_len));

  __bro_io_msg_queue_flush(bc);

  /* Check that the queue does not grow too big: */
  while (bc->msg_queue_len > BRO_MSG_QUEUELEN_MAX)
    {
      BroMsg *msg = bc->msg_queue.tqh_first;

      TAILQ_REMOVE(&bc->msg_queue, msg, msg_queue);
      __bro_io_msg_free(msg);
      bc->msg_queue_len--;
      
      D(("Dropped one message due to excess queue length, now %i\n", bc->msg_queue_len));
    }

  D_RETURN_(TRUE);
}


/* Non-static stuff below: ------------------------------------------- */

BroMsg *
__bro_io_msg_new(char type, uint32 peer_id)
{
  static int msg_counter = 0;
  BroMsg *msg;

  if (! (msg = calloc(1, sizeof(BroMsg))))
    return NULL;  
  
  msg->msg_header.hdr_type    = type;
  msg->msg_header.hdr_peer_id = peer_id;
  msg->msg_cont_type = BRO_MSG_CONT_NONE;
  msg->msg_num = msg_counter++;

  return msg;
}


void
__bro_io_msg_free(BroMsg *msg)
{
  if (!msg)
    return;

  switch (msg->msg_cont_type)
    {
    case BRO_MSG_CONT_RAW:
      if (msg->msg_cont_raw)
	free(msg->msg_cont_raw);
      break;
      
    case BRO_MSG_CONT_EVENT:
      __bro_event_free(msg->msg_cont_ev);
      break;
      
    case BRO_MSG_CONT_REQUEST:
      __bro_event_request_free(msg->msg_cont_req);
      break;
#ifdef BRO_PCAP_SUPPORT
    case BRO_MSG_CONT_PACKET:
      bro_packet_free(msg->msg_cont_packet);
      break;
#endif

    default:
      break;
    }
  
  free(msg);
}


void     
__bro_io_msg_set_cont(BroMsg *msg, int type, void *content)
{
  if (!msg)
    return;

  msg->msg_cont_type = type;

  switch (type)
    {
    case BRO_MSG_CONT_RAW:
      msg->msg_cont_raw = (BroBuf*) content;
      D(("Setting raw content for message, type now %i, buffer data: %p\n",
	 msg->msg_cont_type, content));
      break;
      
    case BRO_MSG_CONT_EVENT:
      msg->msg_cont_ev = (BroEvent *) content;
      D(("Setting event content for message, type now %i, event: %s\n",
	 msg->msg_cont_type, msg->msg_cont.msg_ev->name.str_val));
      break;
      
    case BRO_MSG_CONT_REQUEST:
      msg->msg_cont_req = (BroRequest *) content;
      D(("Setting request content for message, type now %i\n", msg->msg_cont_type));
      break;

#ifdef BRO_PCAP_SUPPORT
    case BRO_MSG_CONT_PACKET:
      msg->msg_cont_packet = (BroPacket *) content;
      break;
#endif

    default:
      msg->msg_cont_type = BRO_MSG_CONT_NONE;
    }
}


int      
__bro_io_msg_queue_flush(BroConn *bc)
{
  BroMsg *msg;

  if (! bc)
    return -1;
  
  while ( (msg = bc->msg_queue.tqh_first))
    {
      if (! io_msg_write(bc, msg))
	return bc->msg_queue_len;
      
      TAILQ_REMOVE(&bc->msg_queue, msg, msg_queue);
      __bro_io_msg_free(msg);
      bc->msg_queue_len--;
    }
  
  D(("Queue length after flush: %i\n", bc->msg_queue_len));
  return bc->msg_queue_len;
}

void
__bro_io_msg_queue_dump(BroConn *bc, const char *message)
{
  BroMsg *msg;

  printf("%s: connection %p, length %i\n", message, bc, bc->msg_queue_len);

  for (msg = bc->msg_queue.tqh_first; msg; msg = msg->msg_queue.tqe_next)
    printf(" -- %s(%i)\n", msg_type_2_str(msg->msg_header.hdr_type), msg->msg_num);
}

int
__bro_io_raw_queue(BroConn *bc, int type, uchar *data, int data_len)
{
  BroMsg *msg;
  int result;

  D_ENTER;

  if (!bc)
    D_RETURN_(FALSE);
  
  if (! (msg = __bro_io_msg_new(type, 0)))
    D_RETURN_(FALSE);

  if (data_len > 0)
    {
      BroBuf *buf;
      
      if (! (buf = __bro_buf_new()))
	{
	  __bro_io_msg_free(msg);
	  D_RETURN_(FALSE);
	}
      
      __bro_buf_append(buf, data, data_len);
      __bro_io_msg_set_cont(msg, BRO_MSG_CONT_RAW, buf);
    }

  result = io_msg_queue(bc, msg);
  D_RETURN_(result);
}

int
__bro_io_event_queue(BroConn *bc, BroEvent *ev)
{
  BroEvent *ev_copy;
  BroMsg *msg;
  int result;

  D_ENTER;

  if (!bc)
    D_RETURN_(FALSE);
  
  if (! (msg = __bro_io_msg_new(BRO_MSG_SERIAL, 0)))
    D_RETURN_(FALSE);

  if (! (ev_copy = __bro_event_copy(ev)))
    {
      D(("Could not clone event\n"));
      D_RETURN_(FALSE);
    }

  __bro_io_msg_set_cont(msg, BRO_MSG_CONT_EVENT, ev_copy);
  result = io_msg_queue(bc, msg);
  D_RETURN_(result);  
}

int
__bro_io_request_queue(BroConn *bc, BroRequest *req)
{
  BroMsg *msg;
  int result;

  D_ENTER;

  if (!bc)
    D_RETURN_(FALSE);

  if (! (msg = __bro_io_msg_new(BRO_MSG_REQUEST, 0)))
    D_RETURN_(FALSE);

  __bro_io_msg_set_cont(msg, BRO_MSG_CONT_REQUEST, req);
  result = io_msg_queue(bc, msg);
  D_RETURN_(result);
}

#ifdef BRO_PCAP_SUPPORT
int
__bro_io_packet_queue(BroConn *bc, BroPacket *packet)
{
  BroPacket *clone;
  BroMsg *msg;
  int result;

  D_ENTER;

  if (!bc)
    D_RETURN_(FALSE);

  if (! (msg = __bro_io_msg_new(BRO_MSG_SERIAL, 0)))
    D_RETURN_(FALSE);

  if (! (clone = bro_packet_clone(packet)))
    {
      __bro_io_msg_free(msg);
      D_RETURN_(FALSE);
    }
  
  __bro_io_msg_set_cont(msg, BRO_MSG_CONT_PACKET, clone);
  result = io_msg_queue(bc, msg);
  D_RETURN_(result);
}
#endif

int
__bro_io_process_input(BroConn *bc)
{
  uint32          buf_off, chunk_size;
  BroMsgHeader    msg_hdr;
  int             result = FALSE;

  D_ENTER;

  if (bc->rx_dead)
    {
      if (! (bc->conn_flags & BRO_CFLAG_RECONNECT))
	D_RETURN_(FALSE);
      
      D(("Connection dead and we want to reconnect ...\n"));
      
      if (! bro_reconnect(bc))
	D_RETURN_(FALSE);
    }
  
  /* Read all available data into receive buffer. Our socket
   * is nonblocking so if nothing's available we'll be back
   * right away and in that case also return from this function.
   */
  if (io_input_buffer_fill(bc) <= 0)
    D_RETURN_(FALSE);

  /* Try to process as much in the input buffer as we can */
  for ( ; ; )
    {  
      D(("----- Attempting to extract a message\n"));

      /* Get the current offset of the buffer pointer to make
       * sure we can reset to it if things go wrong.
       */
      buf_off = __bro_buf_ptr_tell(&bc->rx_buf);
      
      /* Now check the buffer contents and see if there's enough
       * for us to analyze it. Start with a uint32 for the size
       * of the first chunk, and then the chunk itself.
       */
      if (! io_read_chunk_size(bc, &chunk_size))
	goto reset_return;
      
      if (chunk_size != sizeof(BroMsgHeader))
	{
	  D(("Received chunk should be %i bytes, but is %i\n",
	     sizeof(BroMsgHeader), chunk_size));
	  io_skip_chunk(&bc->rx_buf, buf_off, chunk_size);
	  result = TRUE;
	  continue;
	}
      
      if (! io_read_msg_hdr(bc, &msg_hdr))
	goto reset_return;
      
      switch (msg_hdr.hdr_type)
	{
	case BRO_MSG_REQUEST:
	  {
	    char *tmp = NULL, *tmp2 = NULL;
	        
	    D(("Received MSQ_REQUEST\n"));
	        
	    /* We need to read another chunk, whose data will contain
	     * a sequence of 0-terminated strings, each one being the
	     * name of an event that the peering Bro is interested in.
	     */
	    if (! io_read_chunk_size(bc, &chunk_size))
	      goto reset_return;
	        
	    if (! (tmp = (char *) malloc(chunk_size * sizeof(char))))
	      goto reset_return;
	        
	    if (! __bro_buf_read_data(&bc->rx_buf, tmp, chunk_size))
	      {
		free(tmp);
		goto reset_return;
	      }
	        
	    for (tmp2 = tmp; tmp2 < tmp + chunk_size; tmp2 = tmp2 + strlen(tmp2) + 1)
	      {
		char *key;

		if (__bro_ht_get(bc->ev_mask, tmp2))
		  continue;

		key = strdup(tmp2);
		__bro_ht_add(bc->ev_mask, key, key);
		D(("Will report event '%s'\n", tmp2));
	      }

	    D(("Now reporting %i event(s).\n", __bro_ht_get_size(bc->ev_mask)));
	    free(tmp);
	  }
	  break;
	    
	case BRO_MSG_VERSION:
	  {
	    uchar *data;
	    uint32 proto_version;
	    uint32 data_version;

	    D(("Received MSG_VERSION\n"));

	    /* We need to read another chunk for the raw data.
	     */
	    if (! io_read_chunk_size(bc, &chunk_size))
	      goto reset_return;

	    if (! (data = malloc(sizeof(uchar) * chunk_size)))
	      goto reset_return;

	    if (! __bro_buf_read_data(&bc->rx_buf, data, chunk_size))
	      {
		free(data);
		goto reset_return;
	      }
	        
	    proto_version = ntohl(((uint32 *) data)[0]);
	    data_version  = ntohl(((uint32 *) data)[2]);
	    
	    if (proto_version != BRO_PROTOCOL_VERSION)
	      {
		D(("EEEK -- we speak protocol version %i, peer speeks %i. Aborting.\n",
		   BRO_PROTOCOL_VERSION, proto_version));		
		__bro_openssl_shutdown(bc);
	      } else {
		D(("Protocols compatible, we speak version %i\n", BRO_PROTOCOL_VERSION));
	      }
	    
	    if (data_version != 0 && data_version != BRO_DATA_FORMAT_VERSION)
	      {
		D(("EEEK -- we speak data format version %i, peer speeks %i. Aborting.\n",
		   BRO_DATA_FORMAT_VERSION, data_version));		
		__bro_openssl_shutdown(bc);
	      } else {
		D(("Data formats compatible, we speak version %i\n", BRO_DATA_FORMAT_VERSION));
	      }
	    
	    free(data);
	  }
	  break;
	    
	case BRO_MSG_SERIAL:
	  {
	    uint32 pre_serial;
	        
	    D(("Received MSQ_SERIAL\n"));
	    pre_serial = __bro_buf_ptr_tell(&bc->rx_buf);

	    if (! io_read_chunk_size(bc, &chunk_size))
	      goto reset_return;
	        
	    if (! io_process_serialization(bc))
	      io_skip_chunk(&bc->rx_buf, pre_serial, chunk_size);
	  }
	  break;
	    
	case BRO_MSG_CAPTURE_FILTER:
	  {
	    uint32 pre_capture;

	    D(("Received MSQ_CAPTURE_FILTER\n"));
	    pre_capture = __bro_buf_ptr_tell(&bc->rx_buf);
	        
	    if (! io_read_chunk_size(bc, &chunk_size))
	      goto reset_return;
	        
	    io_skip_chunk(&bc->rx_buf, pre_capture, chunk_size);
	  }
	  break;

	case BRO_MSG_HANDSHAKE_COMPLETE:
	  /* No additional content for this one. */
	  switch (bc->handshake_state)
	    {
	    case BRO_HANDSHAKE_SENT:
	      D(("Handshake state for conn %p is COMPLETE\n", bc));
	      bc->handshake_state = BRO_HANDSHAKE_COMPLETE;
	      break;
	      
	    case BRO_HANDSHAKE_COMPLETE:
	      /* We ignore those */
	      break;
	      
	    default:
	      D(("Handshake state for conn %p is RECEIVED\n", bc));
	      bc->handshake_state = BRO_HANDSHAKE_RECEIVED;
	    }
	  break;
	    
	default:
	  D(("Skipping unknown message type %i\n", msg_hdr.hdr_type));
	  break;
	}
      
      __bro_buf_consume(&bc->rx_buf);
      result = TRUE;
    }
  
 reset_return:
  __bro_buf_ptr_seek(&bc->rx_buf, buf_off, SEEK_SET);
  D_RETURN_(result);
}


int
__bro_io_send_handshake_complete(BroConn *bc)
{
  D_ENTER;

  if (! bc)
    D_RETURN_(FALSE);

  if (bc->handshake_state == BRO_HANDSHAKE_COMPLETE ||
      bc->handshake_state == BRO_HANDSHAKE_SENT)
    D_RETURN_(TRUE);

  if (! __bro_io_raw_queue(bc, BRO_MSG_HANDSHAKE_COMPLETE, NULL, 0))
    {
      D_RETURN_(FALSE);
    }

  switch (bc->handshake_state)
    {
    case BRO_HANDSHAKE_RECEIVED:
      /* Note that anything we send after the handshake-complete
       * message above neccessarily has to be processed by the peer
       * after that message, hence there is no race between the time
       * when we send the message and when it is processed by the peer.
       */
      D(("Handshake state for conn %p is COMPLETE\n", bc));
      bc->handshake_state = BRO_HANDSHAKE_COMPLETE;
      break;

    default:
      D(("Handshake state for conn %p is SENT\n", bc));
      bc->handshake_state = BRO_HANDSHAKE_SENT;
    }

  D_RETURN_(TRUE);
}
