/*
       B R O C C O L I  --  The Bro Client Communications Library

Copyright (C) 2004 Christian Kreibich <christian.kreibich(at)cl.cam.ac.uk>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#ifdef __EMX__
#include <strings.h>
#endif 

#include <bro_types.h>
#include <bro_io.h>
#include <bro_debug.h>
#include <bro_val.h>
#include <bro_record.h>


BroRecord *
__bro_record_new(void)
{
  BroRecord *rec;

  if (! (rec = calloc(1, sizeof(BroRecord))))
    return NULL;

  TAILQ_INIT(& (rec->val_list));

  return rec;
}


void
__bro_record_free(BroRecord *rec)
{
  BroVal *val;

  if (! rec)
    return;

  while ( (val = rec->val_list.tqh_first))
    {
      char *field;
      BroSObject *obj;
      
      TAILQ_REMOVE(&rec->val_list, val, val_list);
      obj = (BroSObject *) val;
      
      if ( (field = __bro_sobject_data_del(obj, "field")))
	free(field);
      
      __bro_sobject_release(obj);
    }
  
  free(rec);
}


BroRecord     *
__bro_record_copy(BroRecord *rec)
{
  BroVal *val, *val_copy;
  BroRecord *copy;

  D_ENTER;
  
  if (! rec)
    D_RETURN_(NULL);
  
  if (! (copy = __bro_record_new()))
    D_RETURN_(NULL);

  for (val = rec->val_list.tqh_first; val; val = val->val_list.tqe_next)
    {
      char *field;

      /* Check if it's an assigned val or not */
      if (! val->val_type)
	{
	  D(("Error: unassigned val in record.\n"));
	  goto error_return;
	}
      
      if (! (val_copy = (BroVal *) __bro_sobject_copy((BroSObject *) val)))
	goto error_return;
      
      if (! val_copy->val_type)
	D(("WARNING -- typeless val duplicated as %p, original %p had type %p\n",
	   val_copy, val, val->val_type));

      if (! (field = __bro_sobject_data_get((BroSObject *) val, "field")))
	{
	  D(("Val %p in record %p doesn't have a field name.\n", val, rec));
	  goto error_return;
	}
            
      __bro_sobject_data_set((BroSObject *) val_copy, "field", strdup(field));      
      __bro_record_add_val(copy, val_copy);
    }
  
  D_RETURN_(copy);
  
 error_return:
  __bro_record_free(copy);
  D_RETURN_(NULL);
}


void
__bro_record_add_val(BroRecord *rec, BroVal *val)
{
  if (! rec || ! val)
    return;
  
  TAILQ_INSERT_TAIL(&rec->val_list, val, val_list);
  rec->val_len++;
}


BroVal *
__bro_record_get_nth_val(BroRecord *rec, int num)
{
  BroVal *val;
  int i;
  
  if (! rec || num < 0 || num >= rec->val_len)
    return NULL;
  
  for (i = 0, val = rec->val_list.tqh_first; i < rec->val_len; i++, val = val->val_list.tqe_next)
    {
      if (i == num)
	return val;
    }
  
  return NULL;
}


BroVal *
__bro_record_get_named_val(BroRecord *rec, const char *name)
{
  BroVal *val;
  
  if (! rec || ! name || ! *name)
    return NULL;
  
  for (val = rec->val_list.tqh_first; val; val = val->val_list.tqe_next)
    {
      char *val_name = __bro_sobject_data_get((BroSObject *) val, "field");
      
      if (val_name && ! strcmp(val_name, name))
	return val;
    }
  
  return NULL;
}


int
__bro_record_set_nth_val(BroRecord *rec, int num, BroVal *v)
{
  int i;
  BroVal *val;

  if (! rec || num < 0 || num >= rec->val_len || ! v)
    return FALSE;

  for (i = 0, val = rec->val_list.tqh_first; val; i++, val = val->val_list.tqe_next)
    {
      if (i == num)
	{
	  /* First insert after the current element ... */
	  TAILQ_INSERT_AFTER(&rec->val_list, val, v, val_list);
	  
	  /* ... and then remove the current one. */
	  TAILQ_REMOVE(&rec->val_list, val, val_list);
	  
	  /* Release the old value. */
	  __bro_sobject_release((BroSObject *) val);
	  
	  return TRUE;
	}
    }
  
  return FALSE;
}


int
__bro_record_set_named_val(BroRecord *rec, const char *name, BroVal *v)
{
  BroVal *val;
  
  if (! rec || ! name || !*name || ! v)
    return FALSE;
  
  for (val = rec->val_list.tqh_first; val; val = val->val_list.tqe_next)
    {
      char *val_name = __bro_sobject_data_get((BroSObject *) val, "field");
      
      if (val_name && ! strcmp(val_name, name))
	{
	  /* First insert after the current element ... */
	  TAILQ_INSERT_AFTER(&rec->val_list, val, v, val_list);
	  
	  /* ... and then remove the current one. */
	  TAILQ_REMOVE(&rec->val_list, val, val_list);
	  
	  /* Release the old value. */
	  __bro_sobject_release((BroSObject *) val);
	  
	  return TRUE;
	}
    }
  
  return FALSE;
}
